package org.glite.authz.pep.obligation.eesmap;
import org.glite.authz.pep.obligation.AbstractObligationHandler;
import java.net.URLEncoder;
import org.opensaml.Configuration;
import org.opensaml.saml2.core.Assertion;
import org.opensaml.saml2.core.Statement;
import org.opensaml.ws.soap.client.http.HttpClientBuilder;
import org.opensaml.ws.soap.client.http.HttpSOAPClient;
import org.opensaml.xml.parse.BasicParserPool;
import org.opensaml.xml.parse.XMLParserException;
import org.opensaml.xml.util.XMLHelper;
import org.glite.authz.pep.obligation.ObligationProcessingException;
import org.glite.authz.common.profile.AuthorizationProfileConstants;
import org.glite.authz.common.model.Attribute;
import org.glite.authz.common.model.Environment;
import org.glite.authz.common.model.Obligation;
import org.glite.authz.common.model.Request;
import org.glite.authz.common.model.Response;
import org.glite.authz.common.model.Result;
import org.glite.authz.common.model.util.XACMLConverter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.opensaml.ws.soap.client.BasicSOAPMessageContext;
import org.glite.authz.common.AuthzServiceConstants;
import org.glite.authz.pep.server.config.PEPDaemonConfiguration;
import org.opensaml.ws.soap.soap11.Envelope;
import org.glite.authz.pep.server.PEPDaemonRequestHandler;
import org.opensaml.xacml.ctx.RequestType;
import org.opensaml.xacml.ctx.ResponseType;
import org.opensaml.xacml.profile.saml.XACMLAuthzDecisionStatementType;
import org.apache.commons.httpclient.methods.PostMethod;
import org.opensaml.xml.io.Unmarshaller;
import org.opensaml.xml.io.UnmarshallingException;
import org.opensaml.ws.soap.client.SOAPClientException;
import org.w3c.dom.Element;
import org.glite.authz.common.model.AttributeAssignment;
import java.io.IOException;
import java.io.InputStream;
import java.util.Set;
import java.util.List;
import java.util.ListIterator;
import java.util.HashSet;
import java.lang.reflect.Method;
import org.opensaml.ws.soap.client.http.HttpSOAPRequestParameters;


public class EESObligationHandler extends AbstractObligationHandler {
	/** Class logger. */
    private final Logger log = LoggerFactory.getLogger(EESObligationHandler.class);
	
    /** DN/FQAN to POSIX account mapper. */
    private PEPDaemonConfiguration 	daemonConfig;
    private PEPDaemonRequestHandler requestHandler;
    private String					endPoint;
    private String					issuer;
    private String					special_ees_obligation_list_attr_id;
    
	public EESObligationHandler(String endPoint) throws IllegalArgumentException {
		/* It's quite possible we will have to change this identifier */
		super(AuthorizationProfileConstants.ID_OBLIGATION_LOCAL_ENV_MAP);
		/* some way to get these statically? */
		this.daemonConfig = new PEPDaemonConfiguration();
		this.requestHandler = new PEPDaemonRequestHandler(daemonConfig);
		/* Aim this at a running EES instance */
		this.endPoint = endPoint;
		this.issuer = "PEPd EES OH";
		this.special_ees_obligation_list_attr_id = "http://authz-interop.org/xacml/attribute/trans-obligation";
		log.info("Loading EES OH!");
	}
	
	public Request mixInPDPObligations(Request request, Result result){
		Environment forwardedEnvironment = request.getEnvironment();
		ListIterator<Obligation> obligations_iterator = result.getObligations().listIterator();
		ListIterator<AttributeAssignment> aa_iterator = null;
		Set<Attribute> newEnvironmentAttributes = new HashSet<Attribute>();
		Obligation current_obligation = null;
		String trans_obligation_name = null;
 
		try {
			while(obligations_iterator.hasNext()){
				Attribute transfer_obligations_attribute = new Attribute();
				current_obligation = obligations_iterator.next();
				aa_iterator = current_obligation.getAttributeAssignments().listIterator();
				
				transfer_obligations_attribute.setId(this.special_ees_obligation_list_attr_id);
				transfer_obligations_attribute.getValues().add(current_obligation.getId());
				transfer_obligations_attribute.setIssuer(this.issuer);
				newEnvironmentAttributes.add(transfer_obligations_attribute);
		
				while(aa_iterator.hasNext()){
					AttributeAssignment aa = aa_iterator.next();
					Attribute a = new Attribute();
					trans_obligation_name = URLEncoder.encode(current_obligation.getId(), "UTF-8") + "?attrId=" + URLEncoder.encode(aa.getAttributeId(), "UTF-8");
					a.setId(trans_obligation_name);
					a.setDataType(aa.getDataType());
					a.setIssuer(this.issuer);
					a.getValues().add(aa.getValue());
					newEnvironmentAttributes.add(a);
				}
			}
			forwardedEnvironment.getAttributes().addAll(newEnvironmentAttributes);
			request.setEnvironment(forwardedEnvironment);
		} catch(Exception e){
			log.error("EES ObligationHandler could not add in PDP obligations into PEP->EES request!");
			e.printStackTrace();
		}
		
		return request;
	}
    
	/* evaluates an incoming Request (from the PEP-client),
	   which is merged with a Result (from a PDP instance)
	   by forwarding it to the EES and interpreting its response */
	public boolean evaluateObligation(Request request, Result result) throws ObligationProcessingException{
		boolean applied= false;
		BasicSOAPMessageContext messageContext = buildMessageContext();
		Envelope inboundMessage, outboundMessage;
		Response authzResponse;
		Result ees_result;
		RequestType request_xacml;
	
		try {
			// Send request with mixed in obligations to EES
			request = mixInPDPObligations(request, result);
			request_xacml = XACMLConverter.requestToXACML(request);
			outboundMessage = buildSOAPMessage(messageContext, request_xacml);
        	messageContext.setOutboundMessage(outboundMessage);
			send(messageContext);
			
			// Receive list of remaining obligations
			try {
				
	        	inboundMessage = (Envelope) messageContext.getInboundMessage();
	        	authzResponse = extractResponse(inboundMessage);
	        	ees_result = authzResponse.getResults().get(0);
	        	result.getObligations().clear();
	            result.getObligations().addAll(ees_result.getObligations());
			} catch (Exception e){
				log.error("Couldn't read EES response.");
			}

            applied = true;
		} catch(Exception e){
			log.error("EES ObligationHandler could not evaluate obligation!");
		}
        return applied;
    }
	
	private BasicSOAPMessageContext buildMessageContext() {
		BasicSOAPMessageContext messageContext = null;
		/* TODO:
		 * XACML authorization request message handling is present in the
		 * PEP daemon, but not public. This is a hack to get it working,
		 * and for the next major release after 1.3.0 a real fix is required.
		 * This means communicating with the pep.server developers the need
		 * for public methods, or else creating a separate class implementing
		 * XACML authorization request functionality that can be used here.
		 */
		try {
			Method m = requestHandler.getClass().getDeclaredMethod("buildMessageContext", new Class[]{});
			m.setAccessible(true);
			messageContext = (BasicSOAPMessageContext)m.invoke(requestHandler, new Object[]{});
		} catch (Exception e) {
			log.error("EES PEPD AuthzRequestContext hack trouble (buildMessageContext): ", e);
			return null;
		}
		// override message issuer from possibly different daemonConfig
		messageContext.setOutboundMessageIssuer(daemonConfig.getEntityId());
	
		// TODO fill in security policy resolver
		return messageContext;
	}  
	
	private Envelope buildSOAPMessage(BasicSOAPMessageContext messageContext, RequestType authzRequest) {
		/* see comment in buildMessageContext() */
		try {
			Class<?> cls = Class.forName("org.glite.authz.pep.server.PEPDaemonRequestHandler$AuthzRequestContext");
			Method m = requestHandler.getClass().getDeclaredMethod("buildSOAPMessage", new Class[]{cls, RequestType.class});
			m.setAccessible(true);
			return (Envelope)m.invoke(requestHandler, new Object[]{ messageContext, authzRequest });
		} catch (Exception e) {
			log.error("EES PEPD AuthzRequestContext hack trouble (buildSOAPMessage: ",e );
		}
		return null;
	}

	/* extracts (Argus) Response object from (OpenSAML) SOAP Envelope */
	private Response extractResponse(Envelope soapResponse) {
		org.opensaml.saml2.core.Response samlResponse = null;
		Assertion samlAssertion  = null;
		List<Statement> authzStatements = null;
		RequestType request = null;
		ResponseType response = null;
		XACMLAuthzDecisionStatementType authzStatement = null;
		
		if(soapResponse != null){
			samlResponse = (org.opensaml.saml2.core.Response) soapResponse.getBody().getOrderedChildren().get(0);

			if(samlResponse != null){
				samlAssertion = samlResponse.getAssertions().get(0);
				authzStatements = samlAssertion.getStatements(XACMLAuthzDecisionStatementType.TYPE_NAME_XACML20);
		
				if(authzStatements != null){
					authzStatement = (XACMLAuthzDecisionStatementType) authzStatements.get(0);
					response = authzStatement.getResponse();
					request = authzStatement.getRequest();
					
					if(request == null){
						log.error("Unable to get request from authzStatement");
						return null;
					}
					
					if(response == null){
						log.error("Unable to get response from authzStatement");
						return null;
					}
					
					return XACMLConverter.responseFromXACML(response, request);
				} else {
					log.error("AuthzStatement is blank!");
				}
			} else {
				log.error("SAML response is blank!");
			}
		} else {
			log.error("SOAP response is blank!");
		}
		
		return null;
	}
	
    /* Sends AuthzRequestContext to endpoint via SOAP */       
    public boolean send(BasicSOAPMessageContext messageContext){
	   BasicParserPool parserPool = new BasicParserPool();
       parserPool.setNamespaceAware(true);
       
       HttpClientBuilder clientBuilder = new HttpClientBuilder();
       log.debug("Starting transfer to EES\n");
       HttpSOAPClient soapClient = new HttpSOAPClient(clientBuilder.buildClient(), parserPool);

       // Send the message
       try {
    	   soapClient.send(endPoint, messageContext);
       } catch (Exception e) {
           log.error("Error sending SOAP message to the EES at {}. Error: {}", endPoint, e.toString());
	   }
	   return true;
	}

    /* Process a response from the data returned via HTTP */
    public void processSuccessfulResponse(PostMethod httpMethod, BasicSOAPMessageContext messageContext)
    	throws SOAPClientException {
    	try {
    		Envelope response = unmarshallResponse(httpMethod.getResponseBodyAsStream());
    		messageContext.setInboundMessage(response);
    		//evaluateSecurityPolicy(messageContext);
    	} catch (IOException e) {
    		throw new SOAPClientException("Unable to read response", e);
    	}
    }

    /* Extracts the (OpenSAML) SOAP Envelope from the received XACML data */
    public Envelope unmarshallResponse(InputStream responseStream) throws SOAPClientException {
    	try {
    		BasicParserPool parserPool = new BasicParserPool();
    		Element responseElem = parserPool.parse(responseStream).getDocumentElement();
    		if (log.isDebugEnabled()) {
    			log.debug("Inbound SOAP message was:\n" + XMLHelper.prettyPrintXML(responseElem));
    		}
    		Unmarshaller unmarshaller = Configuration.getUnmarshallerFactory().getUnmarshaller(responseElem);
    		return (Envelope) unmarshaller.unmarshall(responseElem);
    	} catch (XMLParserException e) {
    		throw new SOAPClientException("Unable to parse the XML within the response", e);
    	} catch (UnmarshallingException e) {
    		throw new SOAPClientException("unable to unmarshall the response DOM", e);
    	}
	}

}
